/* rap~ - resonant all-pass filter */

#include "m_pd.h"
#include <math.h>

/* Claude Heiland's fast cosine function from code.mathr.co.uk */

#define   likely(x) __builtin_expect((x),1)
#define unlikely(x) __builtin_expect((x),0)

// postcondition: 0 <= phase <= 1
static inline float cosf9_reduce(float phase) {
  float p = fabsf(phase);
  // p >= 0
  if (likely(p < (1 << 24))) {
    int q = p;
    return p - q;
  } else {
    if (unlikely(isnanf(p) || isinff(p))) {
      // return NaN
      return p - p;
    } else {
      // int could overflow, and it will be integral anyway
      return 0.0f;
    }
  }
}

// precondition: 0 <= phase <= 1
static inline float cosf9_unsafe(float phase) {
  float p = fabsf(4.0f * phase - 2.0f) - 1.0f;
  // p in -1 .. 1
  float s
    = 1.5707963267948965580e+00f * p
    - 6.4596271553942852250e-01f * p * p * p
    + 7.9685048314861006702e-02f * p * p * p * p * p
    - 4.6672571910271187789e-03f * p * p * p * p * p * p * p
    + 1.4859762069630022552e-04f * p * p * p * p * p * p * p * p * p;
  // compiler figures out optimal simd multiplications
  return s;
}

static inline float cosf9(float phase) {
  return cosf9_unsafe(cosf9_reduce(phase));
}

extern void cosf9s1(int n, const float * in, float *  out) {
  while (n--) {
    *out++ = cosf9(*in++);
  }
}

extern void cosf9s2(int n, const float *  in, float *  out,
    float *  tmp) {
  float *  tmp2 = tmp;
  int m = n;
  while (n--) {
    *tmp++ = cosf9_reduce(*in++);
  }
  while (m--) {
    *out++ = cosf9_unsafe(*tmp2++);
  }
}

typedef struct _rap
{
    t_object x_obj;
    t_float x_f;
    t_float x_state1;
    t_float x_state2;
    t_float x_state3;
    t_float x_isr;
} t_rap;

static t_class *rap_class;

static void *rap_new(void)
{
    t_rap *x = (t_rap *)pd_new(rap_class);
    inlet_new(&x->x_obj, &x->x_obj.ob_pd, &s_signal, &s_signal);
    inlet_new(&x->x_obj, &x->x_obj.ob_pd, &s_signal, &s_signal);
    outlet_new(&x->x_obj, gensym("signal"));
    x->x_f = x->x_state1 = x->x_state2 = 0;
    return (x);
}

static t_int *rap_perform(t_int *w)
{
    t_rap *x = (t_rap *)(w[1]);
    t_float *in1 = (t_float *)(w[2]);
    t_float *in2 = (t_float *)(w[3]);
    t_float *in3 = (t_float *)(w[4]);
    t_float *out = (t_float *)(w[5]);
    int n = (int)(w[6]), i;
    t_float state1 = x->x_state1;
    t_float state2 = x->x_state2, tmp, fixamp;
    if (PD_BIGORSMALL(state1))
        state1 = 0;
    if (PD_BIGORSMALL(state2))
        state2 = 0;
    for (i = 0; i < n; i++)
    {
        t_float resonance = *in2++ * x->x_isr, cosres, sinres;
        t_float mixing = *in3++ * x->x_isr, cosmix, sinmix;
        t_float tmp1, intmp;
        cosres = cosf9_reduce(resonance);
        sinres = cosf9_reduce(0.25-resonance);
        cosmix = cosf9_reduce(mixing);
        sinmix = cosf9_reduce(0.25-mixing);
        cosres = cosf9_unsafe(cosres);
        sinres = cosf9_unsafe(sinres);
        cosmix = cosf9_unsafe(cosmix);
        sinmix = cosf9_unsafe(sinmix);
        fixamp = 1.5f - 0.5f*(cosres*cosres + sinres*sinres);
        cosres *= fixamp;
        sinres *= fixamp;
        fixamp = 1.5f - 0.5f*(cosmix*cosmix + sinmix*sinmix);
        cosmix *= fixamp;
        sinmix *= fixamp;
        intmp = *in1++;
        tmp = cosres * state1 - sinres * state2;
        state2 = cosres * state2 + sinres * state1;
        state1 = cosmix * tmp + sinmix * intmp;
        *out++ = cosmix * intmp - sinmix * tmp;
    }
    x->x_state1 = state1;
    x->x_state2 = state2;
    return (w+7);
}

static void rap_dsp(t_rap *x, t_signal **sp)
{
    x->x_isr = 1/sp[0]->s_sr;
    dsp_add(rap_perform, 6, x, sp[0]->s_vec, sp[1]->s_vec,
        sp[2]->s_vec, sp[3]->s_vec, sp[0]->s_n);
}

void rap_tilde_setup(void)
{
    int i;
    rap_class = class_new(gensym("rap~"),
        (t_newmethod)rap_new, 0, sizeof(t_rap), 0, 0);
    class_addmethod(rap_class, (t_method)rap_dsp, gensym("dsp"), A_CANT, 0);
    class_addmethod(rap_class, (t_method)rap_dsp, gensym("signal"),
    	A_CANT, 0);
    CLASS_MAINSIGNALIN(rap_class, t_rap, x_f);
}
